import asyncio
import collections
import logging
import re
import struct
import sys
from time import time

# from tornado import ioloop, iostream, gen, concurrent, tcpclient

from aiozk import protocol, iterables, exc

DEFAULT_READ_TIMEOUT = 3

version_regex = re.compile(rb'Zookeeper version: (\d+)\.(\d+)\.(\d+)-.*')

# all requests and responses are prefixed with a 32-bit int denoting size
size_struct = struct.Struct("!i")
# replies are prefixed with an xid, zxid and error code
reply_header_struct = struct.Struct("!iqi")

log = logging.getLogger(__name__)
payload_log = logging.getLogger(__name__ + ".payload")
if payload_log.level == logging.NOTSET:
    payload_log.setLevel(logging.INFO)


class Connection:

    def __init__(self, host, port, watch_handler, read_timeout, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.host = host
        self.port = int(port)

        self.reader = None
        self.writer = None
        self.closing = False

        self.version_info = None
        self.start_read_only = None

        self.watch_handler = watch_handler

        self.opcode_xref = {}
        self.host_ip = None

        self.pending = {}
        self.pending_specials = collections.defaultdict(list)

        self.watches = collections.defaultdict(list)

        self.read_timeout = read_timeout or DEFAULT_READ_TIMEOUT
        self._read_loop = None

    async def connect(self):
        log.debug("Initial connection to server %s:%d", self.host, self.port)
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port,
                                                                 loop=self.loop)

        self.host_ip = self.writer.transport.get_extra_info('peername')[0]

        log.debug("Sending 'srvr' command to %s:%d", self.host, self.port)
        self.writer.write(b"srvr")

        answer = await self.reader.read()

        version_line = answer.split(b"\n")[0]
        match = version_regex.match(version_line)
        if match is None:
            raise ConnectionError
        self.version_info = tuple(map(int, match.groups()))
        self.start_read_only = bool(b"READ_ONLY" in answer)

        log.debug("Version info: %s", self.version_info)
        log.debug("Read-only mode: %s", self.start_read_only)

        log.debug("Actual connection to server %s:%d", self.host, self.port)
        self.writer.close()
        self.reader, self.writer = await asyncio.open_connection(self.host, self.port,
                                                                 loop=self.loop)

    async def send_connect(self, request):
        # meant to be used before the read_loop starts
        payload_log.debug("[SEND] (initial) %s", request)

        payload = request.serialize()
        payload = size_struct.pack(len(payload)) + payload

        self.writer.write(payload)

        try:
            _, zxid, response = await self.read_response(initial_connect=True)
        except Exception:
            log.exception("Error reading connect response.")
            return

        payload_log.debug("[RECV] (initial) %s", response)
        return zxid, response

    def start_read_loop(self):
        self._read_loop = self.loop.create_task(self.read_loop())
        # ioloop.IOLoop.current().add_callback(self.read_loop)

    def send(self, request, xid=None):
        f = self.loop.create_future()

        if self.closing:
            f.set_exception(exc.ConnectError(self.host, self.port))
            return f

        if request.special_xid:
            xid = request.special_xid

        payload_log.debug("[SEND] (xid: %s) %s", xid, request)

        payload = request.serialize(xid)
        payload = size_struct.pack(len(payload)) + payload

        self.opcode_xref[xid] = request.opcode

        if xid in protocol.SPECIAL_XIDS:
            self.pending_specials[xid].append(f)
        else:
            self.pending[xid] = f

        def handle_write(write_future):
            try:
                write_future.result()
            except Exception as e:
                log.exception('Handle future error')
                self.abort()

        try:
            self.writer.write(payload)
            # handle_write(f)
        except Exception as e:
            log.exception('Exception during write')
            self.abort()

        return f

    async def read_loop(self):
        """
        Infinite loop that reads messages off of the socket while not closed.

        When a message is received its corresponding pending Future is set
        to have the message as its result.

        This is never used directly and is fired as a separate callback on the
        I/O loop via the `connect()` method.
        """
        while not self.closing:
            try:
                xid, zxid, response = await self.read_response()
            except (ConnectionAbortedError, asyncio.CancelledError):
                return
            except Exception as e:
                log.exception("Error reading response.")
                self.abort()
                return

            payload_log.debug("[RECV] (xid: %s) %s", xid, response)

            if xid == protocol.WATCH_XID:
                self.watch_handler(response)
                continue
            elif xid in protocol.SPECIAL_XIDS:
                f = self.pending_specials[xid].pop()
            else:
                f = self.pending.pop(xid)

            if isinstance(response, Exception):
                f.set_exception(response)
            elif not f.cancelled():
                f.set_result((zxid, response))

    async def _read(self, size=-1):
        remaining_size = size
        end_time = time() + self.read_timeout
        payload = []
        while remaining_size and (time() < end_time):
            remaining_time = end_time - time()
            done, pending = await asyncio.wait([self.reader.read(remaining_size)],
                                               timeout=remaining_time,
                                               loop=self.loop)
            if done:
                chunk = done.pop().result()
                payload.append(chunk)
                remaining_size -= len(chunk)
            if pending:
                pending.pop().cancel()
        if remaining_size:
            raise exc.UnfinishedRead
        return b''.join(payload)

    async def read_response(self, initial_connect=False):
        raw_size = await self.reader.read(size_struct.size)
        if raw_size == b'':
            raise ConnectionAbortedError
        size = size_struct.unpack(raw_size)[0]

        # connect and close op replies don't contain a reply header
        if initial_connect or self.pending_specials[protocol.CLOSE_XID]:
            raw_payload = await self._read(size)
            response = protocol.ConnectResponse.deserialize(raw_payload)
            return (None, None, response)

        raw_header = await self._read(reply_header_struct.size)
        xid, zxid, error_code = reply_header_struct.unpack_from(raw_header)

        if error_code:
            return (xid, zxid, exc.get_response_error(error_code))

        size -= reply_header_struct.size

        raw_payload = await self._read(size)

        if xid == protocol.WATCH_XID:
            response = protocol.WatchEvent.deserialize(raw_payload)
        else:
            opcode = self.opcode_xref.pop(xid)
            response = protocol.response_xref[opcode].deserialize(raw_payload)

        return (xid, zxid, response)

    def abort(self, exception=exc.ConnectError):
        """
        Aborts a connection and puts all pending futures into an error state.

        If ``sys.exc_info()`` is set (i.e. this is being called in an exception
        handler) then pending futures will have that exc info set.  Otherwise
        the given ``exception`` parameter is used (defaults to
        ``ConnectError``).
        """
        log.warning("Aborting connection to %s:%s", self.host, self.port)

        def abort_pending(f):
            exc_info = sys.exc_info()
            # TODO
            log.debug('Abort pending: {}'.format(f))
            if False and any(exc_info):
                f.set_exc_info(exc_info)
            else:
                f.set_exception(exception(self.host, self.port))

        for pending in self.drain_all_pending():
            if pending.done() or pending.cancelled():
                continue
            abort_pending(pending)

    def drain_all_pending(self):
        for special_xid in protocol.SPECIAL_XIDS:
            for f in iterables.drain(self.pending_specials[special_xid]):
                yield f
        for _, f in iterables.drain(self.pending):
            yield f

    async def close(self, timeout):
        if self.closing:
            return
        if self.pending or (self.pending_specials and self.pending_specials != {None: []}):
            log.warning('Pendings: {}; specials:  {}'.format(self.pending, self.pending_specials))
        self.closing = True

        try:
            # await list(pending_with_timeouts)
            self.abort(exception=exc.TimeoutError)
            # wlist = list(self.drain_all_pending())
            # log.warning('Wait for list: {} {}'.format(wlist, self.pending))
            # if len(wlist) > 0:
            #     await asyncio.wait(wlist, timeout=timeout)
        except asyncio.TimeoutError:
            log.warning('ABORT Timeout')
            await self.abort(exception=exc.TimeoutError)
        except Exception as e:
            log.exception('in close: {}'.format(e))
            raise e
        finally:
            log.debug('Closing writer')
            self.writer.close()
            log.debug('Writer closed')
        if self._read_loop and not self._read_loop.done():
            self._read_loop.cancel()



__version__ = '0.6.0'
version_info = __version__.split('.')

from .client import ZKClient  # noqa
from .protocol import WatchEvent  # noqa
from .protocol.acl import ACL  # noqa
from .retry import RetryPolicy  # noqa



import asyncio

import pytest


@pytest.mark.asyncio
async def test_barrier(zk, path):
    is_lifted = False
    is_worker_started = asyncio.Future()

    async def start_worker():
        barrier = zk.recipes.Barrier(path)
        is_worker_started.set_result('ok')
        await barrier.wait()
        assert is_lifted == True

    barrier = zk.recipes.Barrier(path)
    await barrier.create()

    worker = asyncio.ensure_future(start_worker())
    is_ok = await is_worker_started
    assert is_ok == 'ok'

    is_lifted = True
    await barrier.lift()
    await worker


@pytest.mark.asyncio
async def test_double_barrier(zk, path):
    num_workers = 0
    workers = []

    async def start_worker(min_workers):
        barrier = zk.recipes.DoubleBarrier(path, min_workers)
        await barrier.enter()
        for i in range(5):
            assert num_workers >= min_workers
        await barrier.leave()

    target = 8
    for _ in range(target):
        num_workers += 1
        workers.append(asyncio.ensure_future(start_worker(target)))
    await asyncio.wait(workers)
    await zk.delete(path)



from .request import Request
from .response import Response
from .part import Part
from .stat import Stat
from .primitives import UString, Int, Vector


class ID(Part):
    """
    """
    parts = (
        ("scheme", UString),
        ("id", UString),
    )


class ACL(Part):
    """
    """
    READ_PERM = 1 << 0
    WRITE_PERM = 1 << 1
    CREATE_PERM = 1 << 2
    DELETE_PERM = 1 << 3
    ADMIN_PERM = 1 << 4

    parts = (
        ("perms", Int),
        ("id", ID),
    )

    @classmethod
    def make(
            cls, scheme, id,
            read=False, write=False, create=False, delete=False, admin=False
    ):
        instance = cls(id=ID(scheme=scheme, id=id))
        instance.set_perms(read, write, create, delete, admin)

        return instance

    def set_perms(self, read, write, create, delete, admin):
        perms = 0
        if read:
            perms |= self.READ_PERM
        if write:
            perms |= self.WRITE_PERM
        if create:
            perms |= self.CREATE_PERM
        if delete:
            perms |= self.DELETE_PERM
        if admin:
            perms |= self.ADMIN_PERM

        self.perms = perms


WORLD_READABLE = ACL.make(
    scheme="world", id="anyone",
    read=True, write=False, create=False, delete=False, admin=False
)

AUTHED_UNRESTRICTED = ACL.make(
    scheme="auth", id="",
    read=True, write=True, create=True, delete=True, admin=True
)

UNRESTRICTED_ACCESS = ACL.make(
    scheme="world", id="anyone",
    read=True, write=True, create=True, delete=True, admin=True
)


class GetACLRequest(Request):
    """
    """
    opcode = 6

    parts = (
        ("path", UString),
    )


class GetACLResponse(Response):
    """
    """
    opcode = 6

    parts = (
        ("acl", Vector.of(ACL)),
        ("stat", Stat),
    )


class SetACLRequest(Request):
    """
    """
    opcode = 7

    parts = (
        ("path", UString),
        ("acl", Vector.of(ACL)),
        ("version", Int),
    )


class SetACLResponse(Response):
    """
    """
    opcode = 7

    parts = (
        ("stat", Stat),
    )



import logging
import asyncio

from aiozk import exc

from .data_watcher import DataWatcher
from .recipe import Recipe


log = logging.getLogger(__name__)


class Counter(Recipe):

    sub_recipes = {
        "watcher": DataWatcher
    }

    def __init__(self, base_path, use_float=False):
        super(Counter, self).__init__(base_path)

        self.value = None

        if use_float:
            self.numeric_type = float
        else:
            self.numeric_type = int

        self.value_sync = self.client.loop.create_future()

    async def start(self):
        self.watcher.add_callback(self.base_path, self.data_callback)
        await asyncio.sleep(0.01, loop=self.client.loop)

        await self.ensure_path()

        raw_value = await self.client.get_data(self.base_path)
        self.value = self.numeric_type(raw_value or 0)

    def data_callback(self, new_value):
        self.value = self.numeric_type(new_value)
        if not self.value_sync.done():
            self.value_sync.set_result(None)
            self.value_sync = self.client.loop.create_future()

    async def set_value(self, value, force=True):
        data = str(value)
        await self.client.set_data(self.base_path, data, force=force)
        log.debug("Set value to '%s': successful", data)
        await self.value_sync

    async def apply_operation(self, operation):
        success = False
        while not success:
            data = str(operation(self.value))
            try:
                await self.client.set_data(self.base_path, data, force=False)
                log.debug("Operation '%s': successful", operation.__name__)
                await self.value_sync
                success = True
            except exc.BadVersion:
                log.debug(
                    "Operation '%s': version mismatch, retrying",
                    operation.__name__
                )
                await self.value_sync

    async def incr(self):

        def increment(value):
            return value + 1

        await self.apply_operation(increment)

    async def decr(self):

        def decrement(value):
            return value - 1

        await self.apply_operation(decrement)

    def stop(self):
        self.watcher.remove_callback(self.base_path, self.data_callback)



import asyncio
import collections
import logging
import time

# from tornado import gen

from aiozk import exc


log = logging.getLogger(__name__)


class RetryPolicy(object):

    def __init__(self, try_limit, sleep_func):
        self.try_limit = try_limit
        self.sleep_func = sleep_func

        self.timings = collections.defaultdict(list)

    async def enforce(self, request=None, loop=None):
        self.timings[id(request)].append(time.time())

        tries = len(self.timings[id(request)])
        if tries == 1:
            return

        if self.try_limit is not None and tries >= self.try_limit:
            raise exc.FailedRetry

        wait_time = self.sleep_func(self.timings[id(request)])
        if wait_time is None or wait_time == 0:
            return
        elif wait_time < 0:
            raise exc.FailedRetry

        log.debug("Waiting %d seconds until next try.", wait_time)
        loop = loop or asyncio.get_event_loop()
        await asyncio.sleep(wait_time, loop=loop)

    def clear(self, request):
        self.timings.pop(id(request), None)

    @classmethod
    def once(cls):
        return cls.n_times(1)

    @classmethod
    def n_times(cls, n):

        def never_wait(_):
            return None

        return cls(try_limit=n, sleep_func=never_wait)

    @classmethod
    def forever(cls):

        def never_wait(_):
            return None

        return cls(try_limit=None, sleep_func=never_wait)

    @classmethod
    def exponential_backoff(cls, base=2, maximum=None):

        def exponential(timings):
            wait_time = base ** len(timings)
            if maximum is not None:
                wait_time = min(maximum, wait_time)

            return wait_time

        return cls(try_limit=None, sleep_func=exponential)

    @classmethod
    def until_elapsed(cls, timeout):

        def elapsed_time(timings):
            if timings:
                first_timing = timings[0]
            else:
                first_timing = time.time()

            return (first_timing + timeout) - time.time()

        return cls(try_limit=None, sleep_func=elapsed_time)



import logging
import asyncio
# from tornado import gen, concurrent

from aiozk import protocol, exc

from .recipes.proxy import RecipeProxy
from .session import Session
from .transaction import Transaction
from .features import Features


log = logging.getLogger(__name__)


class ZKClient(object):

    def __init__(
            self,
            servers,
            chroot=None,
            session_timeout=10,
            default_acl=None,
            retry_policy=None,
            allow_read_only=False,
            read_timeout=None,
            loop=None
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.chroot = None
        if chroot:
            self.chroot = self.normalize_path(chroot)
            log.info("Using chroot '%s'", self.chroot)

        self.session = Session(
            servers, session_timeout, retry_policy, allow_read_only, read_timeout, loop
        )

        self.default_acl = default_acl or [protocol.UNRESTRICTED_ACCESS]

        self.stat_cache = {}

        self.recipes = RecipeProxy(self)

    def normalize_path(self, path):
        if self.chroot:
            path = "/".join([self.chroot, path])

        normalized = "/".join([
            name for name in path.split("/")
            if name
        ])

        return "/" + normalized

    def denormalize_path(self, path):
        if self.chroot and path.startswith(self.chroot):
            path = path[len(self.chroot):]

        return path

    async def start(self):
        await self.session.start()

        if self.chroot:
            await self.ensure_path("/")

    @property
    def features(self):
        if self.session.conn:
            return Features(self.session.conn.version_info)
        else:
            return Features((0, 0, 0))

    async def send(self, request):
        response = await self.session.send(request)

        if getattr(request, "path", None) and getattr(response, "stat", None):
            self.stat_cache[
                self.denormalize_path(request.path)
            ] = response.stat

        return response

    async def close(self):
        await self.session.close()

    def wait_for_events(self, event_types, path):
        path = self.normalize_path(path)

        f = self.loop.create_future()

        def set_future(_):
            if not f.done():
                f.set_result(None)
            for event_type in event_types:
                self.session.remove_watch_callback(event_type, path, set_future)

        for event_type in event_types:
            self.session.add_watch_callback(event_type, path, set_future)

        return f

    async def exists(self, path, watch=False):
        path = self.normalize_path(path)

        try:
            await self.send(protocol.ExistsRequest(path=path, watch=watch))
        except exc.NoNode:
            return False
        return True

    async def create(
            self, path, data=None, acl=None,
            ephemeral=False, sequential=False, container=False,
    ):
        if container and not self.features.containers:
            raise ValueError("Cannot create container, feature unavailable.")

        path = self.normalize_path(path)
        acl = acl or self.default_acl

        if self.features.create_with_stat:
            request_class = protocol.Create2Request
        else:
            request_class = protocol.CreateRequest

        request = request_class(path=path, data=data, acl=acl)
        request.set_flags(ephemeral, sequential, container)

        response = await self.send(request)

        return self.denormalize_path(response.path)

    async def ensure_path(self, path, acl=None):
        path = self.normalize_path(path)

        acl = acl or self.default_acl

        paths_to_make = []
        for segment in path[1:].split("/"):
            if not paths_to_make:
                paths_to_make.append("/" + segment)
                continue

            paths_to_make.append("/".join([paths_to_make[-1], segment]))

        while paths_to_make:
            path = paths_to_make[0]

            if self.features.create_with_stat:
                request = protocol.Create2Request(path=path, acl=acl)
            else:
                request = protocol.CreateRequest(path=path, acl=acl)
            request.set_flags(
                ephemeral=False, sequential=False,
                container=self.features.containers
            )

            try:
                await self.send(request)
            except exc.NodeExists:
                pass

            paths_to_make.pop(0)

    async def delete(self, path, force=False):
        path = self.normalize_path(path)

        if not force and path in self.stat_cache:
            version = self.stat_cache[path].version
        else:
            version = -1

        await self.send(protocol.DeleteRequest(path=path, version=version))

    async def deleteall(self, path):
        childs = await self.get_children(path)
        for child in childs:
            await self.deleteall('/'.join([path, child]))
        await self.delete(path, force=True)

    async def get_data(self, path, watch=False):
        path = self.normalize_path(path)

        response = await self.send(
            protocol.GetDataRequest(path=path, watch=watch)
        )
        return response.data

    async def set_data(self, path, data, force=False):
        path = self.normalize_path(path)

        if not force and path in self.stat_cache:
            version = self.stat_cache[path].version
        else:
            version = -1

        await self.send(
            protocol.SetDataRequest(path=path, data=data, version=version)
        )

    async def get_children(self, path, watch=False):
        path = self.normalize_path(path)

        response = await self.send(
            protocol.GetChildren2Request(path=path, watch=watch)
        )
        return response.children

    async def get_acl(self, path):
        path = self.normalize_path(path)

        response = await self.send(protocol.GetACLRequest(path=path))
        return response.acl

    async def set_acl(self, path, acl, force=False):
        path = self.normalize_path(path)

        if not force and path in self.stat_cache:
            version = self.stat_cache[path].version
        else:
            version = -1

        await self.send(
            protocol.SetACLRequest(path=path, acl=acl, version=version)
        )

    def begin_transaction(self):
        return Transaction(self)



import asyncio
import collections
import itertools
import json


from .data_watcher import DataWatcher
from .party import Party
from .lock import Lock
from .recipe import Recipe


class Allocator(Recipe):

    sub_recipes = {
        "party": (Party, ["member_path", "name"]),
        "lock": (Lock, ["lock_path"]),
        "data_watcher": DataWatcher,
    }

    def __init__(self, base_path, name, allocator_fn=None):
        self.name = name

        super(Allocator, self).__init__(base_path)

        if allocator_fn is None:
            allocator_fn = round_robin

        self.allocator_fn = allocator_fn

        self.active = False

        self.full_allocation = collections.defaultdict(set)
        self.full_set = set()

    @property
    def lock_path(self):
        return self.base_path + "/lock"

    @property
    def member_path(self):
        return self.base_path + "/members"

    @property
    def allocation(self):
        return self.full_allocation[self.name]

    def validate(self, new_allocation):
        as_list = []
        for subset in new_allocation.values():
            as_list.extend(list(subset))

        # make sure there are no duplicates among the subsets
        assert len(as_list) == len(set(as_list)), (
            "duplicate items found in allocation: %s" % self.full_allocation
        )
        # make sure there's no mismatch beween the full set and allocations
        assert len(self.full_set.symmetric_difference(set(as_list))) == 0, (
            "mismatch between full set and allocation: %s vs %s" % (
                self.full_set, self.full_allocation
            )
        )

    async def start(self):
        self.active = True

        await self.ensure_path()

        await self.party.join()

        self.data_watcher.add_callback(self.base_path, self.handle_data_change)
        asyncio.ensure_future(self.monitor_member_changes(), loop=self.client.loop)

    async def add(self, new_item):
        new_set = self.full_set.copy().add(new_item)
        await self.update_set(new_set)

    async def remove(self, new_item):
        new_set = self.full_set.copy().remove(new_item)
        await self.update_set(new_set)

    async def update(self, new_items):
        new_items = set(new_items)
        data = json.dumps(list(new_items))

        with (await self.lock.acquire()):
            await self.client.set_data(self.base_path, data=data)

    def monitor_member_changes(self):
        while self.active:
            yield self.party.wait_for_change()
            if not self.active:
                break

            self.allocate()

    def handle_data_change(self, new_set_data):
        if new_set_data is None:
            return

        new_set_data = set(json.loads(new_set_data))
        if new_set_data == self.full_set:
            return

        self.full_set = new_set_data
        self.allocate()

    def allocate(self):
        new_allocation = self.allocator_fn(self.party.members, self.full_set)
        self.validate(new_allocation)
        self.full_allocation = new_allocation

    async def stop(self):
        await self.party.leave()

        self.data_watcher.remove_callback(
            self.base_path, self.handle_data_change
        )


def round_robin(members, items):
    """
    Default allocator with a round robin approach.

    In this algorithm, each member of the group is cycled over and given an
    item until there are no items left.  This assumes roughly equal capacity
    for each member and aims for even distribution of item counts.
    """
    allocation = collections.defaultdict(set)

    for member, item in zip(itertools.cycle(members), items):
        allocation[member].add(item)

    return allocation
